#
# MVFX : Music Visual Effects (tentative name)
#
# zhuo : http://geocities.yahoo.co.jp/gl/zhuoware
#
# Rev. 0.02   April 6, 2008  zhuo
# Rev. 0.05   May 3, 2008  zhuo
# Rev. 0.06   May 5, 2008  zhuo
# Rev. 0.07   May 6, 2008  zhuo
# Rev. 0.08   May 9, 2008  zhuo
#             - loudness scaling now applied in SMF-dump only
#             - vertical cursor introduced
#             - text dump introduced
# Rev. 0.09   May 10, 2008  zhuo 
#             - playback functionality
#                 left press in waveform - hold - release
# Rev. 0.10   May 11, 2008  zhuo
#             - SMF: CC#8 was used by mistake; Now CC#11 is used, correctly. 
#                    scaling for expression adjusted by try-and-error.
# Rev. 0.11   May 13, 2008  zhuo
#             - loudness : windowsize is now wider, to apply smoothing.
#

######################################################################
# package dependencies
######################################################################
package require snack 2.2


######################################################################
# global constants
######################################################################
set MVFXSysAttr(MVFXVersion) "0.11"


######################################################################
# VIEWER SECTION
######################################################################
set MVFXVizAttr(mainW) 2048
set MVFXVizAttr(mainH) 1024
set MVFXVizAttr(mainWinH) 480


set MVFXVizAttr(panelWinW) 160
set MVFXVizAttr(panelLeft) 0
set MVFXVizAttr(panelRight) 480
set MVFXVizAttr(panelTop) 0
set MVFXVizAttr(panelBottom) $MVFXVizAttr(mainH)
set MVFXVizAttr(panelBGColor) black
set MVFXVizAttr(panelFont) {Ariel 9}

set MVFXVizAttr(scaleWinH) 20
set MVFXVizAttr(scaleLeft) 0
set MVFXVizAttr(scaleRight) $MVFXVizAttr(mainW)
set MVFXVizAttr(scaleTop) 0
set MVFXVizAttr(scaleBottom) $MVFXVizAttr(scaleWinH)
set MVFXVizAttr(scaleTickTall) 7
set MVFXVizAttr(scaleTickShort) 4
set MVFXVizAttr(scaleBGColor) "\#404040"
set MVFXVizAttr(scaleFGColor) white
set MVFXVizAttr(scaleFont) {Ariel 8}

set MVFXVizAttr(dispLeft) 0
set MVFXVizAttr(dispTop) 0
set MVFXVizAttr(dispRight) $MVFXVizAttr(mainW)
set MVFXVizAttr(dispBottom) $MVFXVizAttr(mainH)

set MVFXVizAttr(dispWinW) 640
set MVFXVizAttr(dispBackGroundColor) black

set MVFXVizAttr(dispWaveformFillColor) yellow
set MVFXVizAttr(dispWaveformHeightPixel) 96
set MVFXVizAttr(dispWaveformSubsample) 25
set MVFXVizAttr(dispPixelsPerSecond) 250

set MVFXVizAttr(dispSpectrogramColorMap) {}
for { set i 0 } { $i < 256 } { set i [expr $i+1] } {
    lappend MVFXVizAttr(dispSpectrogramColorMap) [format "\#%02x%02x00" $i $i]
}
set MVFXVizAttr(dispSpectrogramHeightPixel) 256
set MVFXVizAttr(dispSpectrogramTopFreq) 1600

## 4096 crashes for a long audio track
#set MVFXVizAttr(dispSpectrogramFFTLength) 4096
#set MVFXVizAttr(dispSpectrogramWindowLength) 4096
set MVFXVizAttr(dispSpectrogramFFTLength) 2048
set MVFXVizAttr(dispSpectrogramWindowLength) 2048
set MVFXVizAttr(dispSpectrogramWindowType) Blackman
set MVFXVizAttr(dispSpectrogramGridColor) yellow

set MVFXVizAttr(dispPitchcurveFrameLength) 0.01
set MVFXVizAttr(dispPitchcurveWindowLength) 0.02
#set MVFXVizAttr(dispPitchcurveWindowLength) 0.0075


#set MVFXVizAttr(dispPitchcurveMaxPitch) 400
#set MVFXVizAttr(dispPitchcurveRefPitch) 110
#set MVFXVizAttr(dispPitchcurveMinPitch) 55
set MVFXVizAttr(dispPitchcurveMaxPitch) 800
set MVFXVizAttr(dispPitchcurveRefPitch) 220 
set MVFXVizAttr(dispPitchcurveMinPitch) 110
set MVFXVizAttr(dispPitchcurveColor) "#ff6060"

set MVFXVizAttr(dispLoudnessColor) "#6060ff"
set MVFXVizAttr(dispLoudnessMIDIThreshold) 48.0
set MVFXVizAttr(dispLoudnessMIDIScaling) 4.5
set MVFXVizAttr(dispLoudnessFrameLength) 0.01
set MVFXVizAttr(dispLoudnessWindowLength) 0.1



set MVFXVizAttr(dispScaleMode) second
#set MVFXVizAttr(dispScaleMode) sample  # not implemented
#set MVFXVizAttr(dispScaleMode) beat    # not implemented


set MVFXVizAttr(dispHorizScaleBGColor) "\#484848"
set MVFXVizAttr(dispHorizScaleHeight) 12

########################################
# 
########################################




########################################
# toplevel window
########################################
set w .top
catch {destroy $w}
toplevel $w
wm title $w "MVFX Rev.$MVFXSysAttr(MVFXVersion)"
wm iconname $w "MVFX Rev.$MVFXSysAttr(MVFXVersion)"
wm withdraw .



########################################
# menu bar
########################################
set menustatus "  "

menu $w.menu -tearoff 0

set mFile $w.menu.file
menu $mFile -tearoff 0
$mFile add command -label "Open Sound File..." -command {MVFXCmdLoadSound $w} -underline 0
$mFile add separator
$mFile add command -label "Exit" -command "destroy $w; exit 0" -underline 1

set mEdit $w.menu.edit
menu $mEdit -tearoff 0
$mEdit add command -label "Play" -command {playbackVsd}

$w.menu add cascade -label "File" -menu $mFile -underline 0
$w.menu add cascade -label "Edit" -menu $mEdit -underline 0
$w configure -menu $w.menu

bind Menu <<MenuSelect>> {
    global $menustatus
    if {[catch {%W entrycget active -label} label]} {
	set label "    "
    }
    set menustatus $label
    update idletasks
}



########################################
# main screen : left-side (panel), right-side(disp)
########################################

#
# tag rules
#
# horizontal scale :    horizScale
# the waveform object :  "snd0" "snd1" ...

frame $w.cframe
canvas $w.cframe.panel \
    -scrollregion \
        "$MVFXVizAttr(panelLeft) $MVFXVizAttr(panelTop) $MVFXVizAttr(panelRight) $MVFXVizAttr(panelBottom)" \
    -width $MVFXVizAttr(panelWinW) \
    -height $MVFXVizAttr(mainWinH) \
    -xscrollcommand "$w.cframe.panelhscroll set" \
    -yscrollcommand "$w.cframe.vscroll set" \
    -background $MVFXVizAttr(panelBGColor) \
    -relief sunken

canvas $w.cframe.scale \
    -width $MVFXVizAttr(mainWinH) \
    -height $MVFXVizAttr(scaleWinH) \
    -xscrollcommand "$w.cframe.disphscroll set" \
    -background $MVFXVizAttr(scaleBGColor) 

canvas $w.cframe.disp \
    -scrollregion "$MVFXVizAttr(dispLeft) $MVFXVizAttr(dispTop) $MVFXVizAttr(dispRight) $MVFXVizAttr(dispBottom)" \
    -width $MVFXVizAttr(dispWinW) \
    -height $MVFXVizAttr(mainWinH) \
    -xscrollcommand "$w.cframe.disphscroll set" \
    -yscrollcommand "$w.cframe.vscroll set" \
    -background $MVFXVizAttr(dispBackGroundColor) \
    -relief sunken

scrollbar $w.cframe.panelhscroll -orient horiz -command "$w.cframe.panel xview"
scrollbar $w.cframe.disphscroll -orient horiz -command "dispHScrollCmd $w"
proc dispHScrollCmd {w args} {
    eval $w.cframe.disp xview $args
    eval $w.cframe.scale xview $args
}
scrollbar $w.cframe.vscroll -command "dispVScrollCmd $w"
proc dispVScrollCmd {w args} {
    eval $w.cframe.panel yview $args
    eval $w.cframe.disp yview $args
}


grid $w.cframe.scale -in $w.cframe \
    -row 0 -column 1 -rowspan 1 -columnspan 1 -sticky news

grid $w.cframe.panel -in $w.cframe \
    -row 1 -column 0 -rowspan 1 -columnspan 1 -sticky news
grid $w.cframe.disp -in $w.cframe \
    -row 1 -column 1 -rowspan 1 -columnspan 1 -sticky news
grid $w.cframe.vscroll \
    -row 1 -column 2 -rowspan 1 -columnspan 1 -sticky news

grid $w.cframe.panelhscroll \
    -row 2 -column 0 -rowspan 1 -columnspan 1 -sticky news
grid $w.cframe.disphscroll \
    -row 2 -column 1 -rowspan 1 -columnspan 1 -sticky news

grid rowconfig $w.cframe 0 -weight 0
grid rowconfig $w.cframe 1 -weight 1
grid rowconfig $w.cframe 2 -weight 0
grid columnconfig $w.cframe 0 -weight 1
grid columnconfig $w.cframe 1 -weight 4
grid columnconfig $w.cframe 2 -weight 0
pack $w.cframe -side top -fill both -expand yes



##############################
# vertical cursor
##############################
bind $w.cframe.disp <Enter> {MVFXCmdActivateVerticalCursor $w %x %y}
bind $w.cframe.disp <Motion> {MVFXCmdUpdateVerticalCursor $w %x %y}
bind $w.cframe.disp <Leave> {MVFXCmdDisactivateVerticalCursor $w %x %y}

proc MVFXCmdActivateVerticalCursor {w x y} { 
    global MVFXVizAttr
    set cx [$w.cframe.disp canvasx $x]
    $w.cframe.disp create line $cx 0 $cx 480 -dash {2 4} -fill "\#c0c0ff" -tags "vertcursor" -state disabled
    set MVFXVizAttr(vertcursor,prevcx) $cx
}

proc MVFXCmdUpdateVerticalCursor {w x y} { 
    global MVFXVizAttr
    set cx [$w.cframe.disp canvasx $x]
    set dx [expr $cx - $MVFXVizAttr(vertcursor,prevcx)]
    $w.cframe.disp move vertcursor $dx 0
    set MVFXVizAttr(vertcursor,prevcx) $cx
}

proc MVFXCmdDisactivateVerticalCursor {w x y} { 
    $w.cframe.disp delete vertcursor
}



##############################
# playback cursor vizualization
##############################

array set MVFXPlaybackCursor {
    isOn 0
    initCx 0
    initClock 0
}

proc MVFXCmdActivatePlaybackCursor {w cx} { 
    global MVFXVizAttr MVFXPlaybackCursor
    $w.cframe.disp create line $cx 0 $cx 480 -dash {4 2} -fill "\#ffc0c0" -tags "playbackcursor" -state disabled
    set MVFXPlaybackCursor(isOn) 1
    set MVFXPlaybackCursor(initCx) $cx
    set MVFXPlaybackCursor(prevCx) $cx
    set MVFXPlaybackCursor(initClock) [clock clicks -milliseconds] 
    after 30 {MVFXCmdUpdatePlaybackCursor $w}
}

proc MVFXCmdUpdatePlaybackCursor {w} { 
    global MVFXVizAttr MVFXPlaybackCursor
    if { $MVFXPlaybackCursor(isOn) == 0 } {
	$w.cframe.disp delete playbackcursor
	return
    }
    set now [clock clicks -milliseconds]
    set cx [expr int( $MVFXPlaybackCursor(initCx) + $MVFXVizAttr(dispPixelsPerSecond) * ($now - $MVFXPlaybackCursor(initClock)) * 0.001 )]
    $w.cframe.disp move playbackcursor [expr $cx - $MVFXPlaybackCursor(prevCx)] 0 
    set MVFXPlaybackCursor(prevCx) $cx
    after 30 {MVFXCmdUpdatePlaybackCursor $w}
}

proc MVFXCmdDisactivatePlaybackCursor {} { 
    global MVFXPlaybackCursor
    set MVFXPlaybackCursor(isOn) 0
}








########################################
# debug console
########################################
if { 1 } {
    text $w.c -width 80 -height 12 -takefocus 0
    pack $w.c -side top -fill both -expand yes
}

proc cPuts {string} {
    global w
    catch {$w.c insert end "$string"}
}


########################################
# dialogs
########################################

proc MVFXDialogChooseSoundFile { w } {
    set types {
	{"Sound Files"	{.wav .mp3 .au .snd .aif .sd .smp .csl} }
    }
    return [tk_getOpenFile -filetypes $types -parent $w]
}





########################################
# visualizer
########################################

proc MVFXVizUpdateMainW { w widthPixel } {
    global MVFXVizAttr 

    set MVFXVizAttr(dispRight) $widthPixel
    catch {$w.cframe.disp configure -scrollregion \
	       "$MVFXVizAttr(dispLeft) $MVFXVizAttr(dispTop) $MVFXVizAttr(dispRight) $MVFXVizAttr(dispBottom)"}
    set MVFXVizAttr(scaleRight) $widthPixel
    catch {$w.cframe.scale configure -scrollregion \
	       "$MVFXVizAttr(scaleLeft) $MVFXVizAttr(scaleTop) $MVFXVizAttr(scaleRight) $MVFXVizAttr(scaleBottom)"}

    MVFXVizSetHorizScale $w
}


proc MVFXVizSetHorizScale { w } {
    global MVFXVizAttr 
    set pixPerSec $MVFXVizAttr(dispPixelsPerSecond)
    set pixMax $MVFXVizAttr(dispRight)
    set tickTallY [expr $MVFXVizAttr(scaleWinH) - $MVFXVizAttr(scaleTickTall)]
    set tickShortY [expr $MVFXVizAttr(scaleWinH) - $MVFXVizAttr(scaleTickShort)]
#    cPuts "$pixMax\n"

    $w.cframe.scale delete all
    for {set t 0; set x 0} {$x <= $pixMax} {set t [expr $t + 1]; set x [expr $x + $pixPerSec]} {
	$w.cframe.scale create text $x $tickTallY -text [format "%3.1f" $t] -justify center -anchor s -font $MVFXVizAttr(scaleFont) -fill $MVFXVizAttr(scaleFGColor)
	$w.cframe.scale create line $x $tickTallY $x $MVFXVizAttr(scaleWinH) -fill $MVFXVizAttr(scaleFGColor) -width 1
	set x2 [expr $x+$pixPerSec/2]
	$w.cframe.scale create line $x2 $tickShortY $x2 $MVFXVizAttr(scaleWinH) -fill $MVFXVizAttr(scaleFGColor) -width 1
    }
}




# add a wave form of the sound (sndHdl) to the screen
# (at the bottom)
proc MVFXVizAddWaveform { w sndHdl } {
    global MVFXSysAttr MVFXSndAttr MVFXVizAttr

    # obtain the width of the waveform
    set waveformWidth [expr int($MVFXSndAttr($sndHdl,sec) * $MVFXVizAttr(dispPixelsPerSecond))]

    # update the display dimension (all scrollable)
    MVFXVizUpdateMainW $w $waveformWidth

    # generate a waveform and put it in
    $w.cframe.disp create waveform 0 0 -anchor nw \
	-fill $MVFXVizAttr(dispWaveformFillColor) \
	-height $MVFXVizAttr(dispWaveformHeightPixel) \
	-pixelspersecond $MVFXVizAttr(dispPixelsPerSecond) \
	-sound $sndHdl \
	-subsample $MVFXVizAttr(dispWaveformSubsample) \
	-tags "$sndHdl $sndHdl,waveform" \
	-frame 1

    ########################################
    # set command menu for this
    ########################################
    menu $w.mWaveform$sndHdl
    $w.mWaveform$sndHdl add command -label "play whole" \
	-command "MVFXSndPlayWhole $w $sndHdl"
    $w.mWaveform$sndHdl add command -label "show spectrogram" \
	-command "MVFXCmdShowSpectrogram $w $sndHdl"
    $w.mWaveform$sndHdl add command -label "show pitch" \
	-command "MVFXCmdShowPitch $w $sndHdl"
    $w.mWaveform$sndHdl add command -label "show loudness" \
	-command "MVFXCmdShowLoudness $w $sndHdl"
    $w.mWaveform$sndHdl add command -label "save SMF" \
	-command "MVFXCmdSaveSMF $w $sndHdl"
    $w.mWaveform$sndHdl add command -label "save Text" \
	-command "MVFXCmdSaveText $w $sndHdl"
    $w.cframe.disp bind "$sndHdl,waveform" <3> "$w.mWaveform$sndHdl post %X %Y"

    ########################################
    # binding for playback
    ########################################
    $w.cframe.disp bind "$sndHdl,waveform" <ButtonPress-1> "MVFXCmdPlayStartFrom $w $sndHdl %x %y"
    $w.cframe.disp bind "$sndHdl,waveform" <ButtonRelease-1> "MVFXCmdPlayStop $w $sndHdl"
    $w.cframe.disp bind "$sndHdl,waveform" <Leave> "MVFXCmdPlayStop $w $sndHdl"
}





proc MVFXVizAddPanel { w sndHdl } {
    global MVFXSysAttr MVFXSndAttr MVFXVizAttr

    # obtain the width of the waveform
    set dirName [file dirname $MVFXSndAttr($sndHdl,filename)]
    set fileName [lindex [file split $MVFXSndAttr($sndHdl,filename)] end]

    # generate a panel for the wave object

    # the coordinate of the wavefile,
    # tentatively set to (0,0)
    set waveformCX 0
    set waveformCY 0
    $w.cframe.panel create text $waveformCX $waveformCY \
	-anchor nw \
	-text "$dirName\n$fileName" \
	-font $MVFXVizAttr(panelFont) \
        -fill $MVFXVizAttr(dispWaveformFillColor) \
	-tags "$sndHdl $sndHdl,panelFileName fname"
    
    set MVFXVizAttr(panelRight) [lindex [$w.cframe.panel bbox fname] 2]
    catch {$w.cframe.panel configure -scrollregion \
	       "$MVFXVizAttr(panelLeft) $MVFXVizAttr(panelTop) $MVFXVizAttr(panelRight) $MVFXVizAttr(panelBottom)"}
}



######################################################################
# CONTROLLER SECTION
######################################################################


proc MVFXCmdLoadSound { w {at "end"} } {
    set fileName [MVFXDialogChooseSoundFile $w]
    if {$fileName == "" } { return }
    $w configure -cursor watch 
    update
    set sndHdl [MVFXSndFromFile $fileName]
    MVFXVizAddWaveform $w $sndHdl
    MVFXVizAddPanel $w $sndHdl
    $w configure -cursor arrow
    update
}


proc MVFXCmdPlayStartFrom { w sndHdl x y } {
    global MVFXVizAttr
    set cx [$w.cframe.disp canvasx $x]
    set startpos [expr int( ($cx - 0.0) * [lindex [$sndHdl info] 1] / $MVFXVizAttr(dispPixelsPerSecond))]
    $sndHdl play -start $startpos -blocking 0 -command "MVFXCmdPlayStop $w $sndHdl"
    MVFXCmdActivatePlaybackCursor $w $cx
}

proc MVFXCmdPlayStop { w sndHdl } {
    $sndHdl stop
    MVFXCmdDisactivatePlaybackCursor
}





######################################################################
# MODEL SECTION
######################################################################


########################################
# sound model management
########################################
set MVFXSndAttr(cnt) 0
set MVFXSndAttr(hdls) {}

# obtain a new sound handle; 
# all the handles are recorded in MVFXSndAttr(hdls) .


proc MVFXSndNew {} {
    global MVFXSndAttr
    set sndHdl "snd$MVFXSndAttr(cnt)"
    incr MVFXSndAttr(cnt)
    lappend MVFXSndAttr(hdls) $sndHdl
    return $sndHdl
}

proc MVFXSndFromFile { fileName } {
    global MVFXSysAttr MVFXSndAttr w
    set sndHdl [MVFXSndNew]
    snack::sound $sndHdl -load $fileName

    # make it mono.
    $sndHdl convert -channels 1
    set MVFXSndAttr($sndHdl,filename) $fileName
    set MVFXSndAttr($sndHdl,sec) [$sndHdl length -unit SECONDS]
    return $sndHdl
}


proc MVFXSndPlayWhole { w sndHdl } {
    $sndHdl play 
# ?-start start? ?-end end? ?-output jack? ?-blocking boolean? -command ?callback? ?-device outputDevice? ?-devicerate rate? ?-devicechannels channels?
}



proc MVFXCmdShowSpectrogram { w sndHdl } {
    global MVFXSysAttr MVFXSndAttr MVFXVizAttr
    $w.cframe.disp create spectrogram \
	0 $MVFXVizAttr(dispWaveformHeightPixel) -anchor nw \
	-fftlength $MVFXVizAttr(dispSpectrogramFFTLength) \
	-colormap $MVFXVizAttr(dispSpectrogramColorMap) \
	-gridcolor $MVFXVizAttr(dispSpectrogramGridColor) \
	-height $MVFXVizAttr(dispSpectrogramHeightPixel) \
	-pixelspersecond $MVFXVizAttr(dispPixelsPerSecond) \
	-topfrequency $MVFXVizAttr(dispSpectrogramTopFreq) \
	-windowtype $MVFXVizAttr(dispSpectrogramWindowType) \
	-winlength $MVFXVizAttr(dispSpectrogramWindowLength) \
	-sound $sndHdl \
	-tags "$sndHdl $sndHdl,spectrogram" 
    $w.cframe.disp create rectangle \
	[$w.cframe.disp bbox $sndHdl,spectrogram] -outline yellow -fill ""\
	-tags "$sndHdl $sndHdl,spectrogram"
    catch {$w.cframe.disp lower $sndHdl,spectrogram "$sndHdl"}
}


proc MVFXCmdShowLoudness { w sndHdl } {
    global MVFXSysAttr MVFXSndAttr MVFXVizAttr

    # params
    set framelength $MVFXVizAttr(dispLoudnessFrameLength)
    set windowLength $MVFXVizAttr(dispLoudnessWindowLength)
    
    # get raw loudness using snack functionality
    set loudnessResult \
	[ $sndHdl power \
	      -framelength $framelength \
	      -windowlength [expr int($windowLength * [lindex [$sndHdl info] 1])] \
	      -windowtype Blackman ]

    set minl 96
    set maxl 0
    set idx 0
    foreach loud $loudnessResult {
	set t [expr $idx * $framelength]
	lappend loudnesscurve $t $loud
	incr idx
    }
    set MVFXSndAttr($sndHdl,loudnesscurve,data) $loudnesscurve
    set MVFXSndAttr($sndHdl,loudnesscurve,fromto) "0.0 $t"

    # visualize the loudnesscurve
    set pps $MVFXVizAttr(dispPixelsPerSecond) 
    set botOfSpec [expr $MVFXVizAttr(dispSpectrogramHeightPixel) + $MVFXVizAttr(dispWaveformHeightPixel)]

    foreach {t l} $loudnesscurve {
	set x [expr int($t * $pps) + 0]
	set y [expr int($botOfSpec - 2.0 * $l)]
	lappend loudnessLine $x $y
    }
    $w.cframe.disp create line $loudnessLine -fill $MVFXVizAttr(dispLoudnessColor) -tags "$sndHdl $sndHdl,loudnesscurve"
    catch {$w.cframe.disp raise $sndHdl,loudnesscurve $sndHdl,spectrogram}
}


proc MVFXCmdShowPitch { w sndHdl } {
    global MVFXSysAttr MVFXSndAttr MVFXVizAttr
    set framelength $MVFXVizAttr(dispPitchcurveFrameLength)

    # get raw pitch result using snack functionality
    set pitchResult \
	[ $sndHdl pitch \
	      -method ESPS \
	      -framelength $framelength \
	      -windowlength $MVFXVizAttr(dispPitchcurveWindowLength) \
	      -maxpitch $MVFXVizAttr(dispPitchcurveMaxPitch) \
	      -minpitch $MVFXVizAttr(dispPitchcurveMinPitch) ]

    # save list in pitchcurve,data; { t0 f0 t1 f1 t2 f2 ... }
    # time begin/end, lowest freq, highest freq  are saved
    # set min, max to its counter (min <= maximum, max <= minimum)
    set minf $MVFXVizAttr(dispPitchcurveMaxPitch)
    set maxf $MVFXVizAttr(dispPitchcurveMinPitch)
    set idx 0
    set f 0
    foreach rec $pitchResult {
	set t [expr $idx * $framelength]
	# simple smoothing 
	set f [expr 0.7 * [lindex $rec 0] + 0.3 * $f]
#	cPuts "($t) ($f)\n"
	lappend pitchcurve $t $f
	if { $f > 0.0 && $f < $minf } { set minf $f }
	if { $f > $maxf } { set maxf $f }
	incr idx
    }
    set MVFXSndAttr($sndHdl,pitchcurve,data) $pitchcurve
    set MVFXSndAttr($sndHdl,pitchcurve,fromtominmax) "0.0 $t $minf $maxf"
    
    cPuts "$MVFXSndAttr($sndHdl,pitchcurve,fromtominmax)\n"

    # visualize the pitchcurve... coord-list in pitchLine
    set pps $MVFXVizAttr(dispPixelsPerSecond) 
    set botOfSpec [expr $MVFXVizAttr(dispSpectrogramHeightPixel) + $MVFXVizAttr(dispWaveformHeightPixel)]

    foreach {t f} $pitchcurve {
	set x [expr int($t * $pps) + 0 ]
	set y [expr $botOfSpec - int( $f * $MVFXVizAttr(dispSpectrogramHeightPixel) / $MVFXVizAttr(dispSpectrogramTopFreq) ) ]
	lappend pitchLine $x $y
    }
    $w.cframe.disp create line $pitchLine -fill $MVFXVizAttr(dispPitchcurveColor) -tags "$sndHdl $sndHdl,pitchcurve"
 
    catch {$w.cframe.disp raise $sndHdl,pitchcurve $sndHdl,spectrogram}
}



# loudness values are NOT scaled; as is, in dB. 96 - 0.
proc MVFXCmdSaveText { w sndHdl } {
    global MVFXSysAttr MVFXSndAttr MVFXVizAttr

    # params
    set types {	{{Text Files}       {.txt}        } }

    # if log-power is below $threshold, it is ignored
    set threshold 45.0
    # expression value is multiplied by this scaling factor
    set expScale 2.5 

    # filename chooser
    set filename [tk_getSaveFile -filetypes $types -defaultextension ".txt"]
    if {"$filename" == ""} {
	return
    }

    # prepare header in tmp text file
    set fd [open $filename "w"]
    set prevPbval ""
    set prevExpval ""

    set pbvalClipped 0
    set expvalClipped 0

    foreach {t f} $MVFXSndAttr($sndHdl,pitchcurve,data) {tdummy l} $MVFXSndAttr($sndHdl,loudnesscurve,data) {
	if { $t == "" || $tdummy == "" } { break }
	puts $fd [format "%8.3f\t%8.3f\t%8.3f" [expr $t * 1000.0] $f $l]
    }
    close $fd
    tk_messageBox -icon info -message "$filename created.\n" -type ok -parent $w
}




#
# expression (loudness  value) is scaled.
proc MVFXCmdSaveSMF { w sndHdl } {
    global MVFXSysAttr MVFXSndAttr MVFXVizAttr

    # params
    set types {	{{Standard MIDI Files}       {.mid}        } }
    set BPM 120
    set TPQ 480
    set meter "4/4"
    set refnote $MVFXVizAttr(dispPitchcurveRefPitch)

    # midi channel is 1-origin in t2mf.
    set midich 1

    # expression value is multiplied by this scaling factor.
    # if log-power is below $threshold, it is ignored.
    set expThreshold $MVFXVizAttr(dispLoudnessMIDIThreshold)
    set expScale $MVFXVizAttr(dispLoudnessMIDIScaling)

    # make sure working files can be created
    set tmptxtfile "T1.TXT"
    set tmpmidfile "T2.MID"
    file delete "$tmpmidfile"
    file delete "$tmptxtfile"

    # filename chooser
    set filename [tk_getSaveFile -filetypes $types -defaultextension ".mid"]
    if {"$filename" == ""} {
	return
    }
    
    # prepare header in tmp text file
    set UsecPerBeat [expr int(60.0* 1000000 / $BPM)]

    set fd [open $tmptxtfile "w"]
    puts $fd "MFile 1 1 480"
    puts $fd "MTrk"
    puts $fd "0 Meta TrkName \"pitchbend\""
    puts $fd "0 Tempo $UsecPerBeat"
    puts $fd "0 TimeSig $meter 24 8"
    puts $fd "0 Par ch=$midich c=101 val=0"
    puts $fd "0 Par ch=$midich c=100 val=0"
    puts $fd "0 Par ch=$midich c=6 val=24"

    set logRefnote [expr log( $refnote )]
    set log2recip [expr 1.0/log(2.0)]
    set prevPbval ""
    set prevExpval ""

    set pbvalClipped 0
    set expvalClipped 0

    foreach {t f} $MVFXSndAttr($sndHdl,pitchcurve,data) {tdummy l} $MVFXSndAttr($sndHdl,loudnesscurve,data) {
	if { $t == "" || $tdummy == "" } { break }
	set tick  [expr int( $t * 1000000 / $UsecPerBeat * $TPQ )]

	# calc pitchbend value for $f
	if { $f == 0.0 } {
	    set pbval 0
	} else {
	    set pbval [expr int( (log( $f ) - $logRefnote) * $log2recip * 4096) ]
	    if { $pbval < -8192 } { set pbval -8192; set pbvalClipped 1 }
	    if { $pbval > 8191 } { set pbval 8191; set pbvalClipped 1 }
	}
	set pbval [expr $pbval + 8192]

	# calc expression value for $l (threshold already applied)
	set expval [expr int( $expScale * ($l - $expThreshold ) )]
	if { $expval < 0 } { set expval 0 }
	if { $expval > 127 } { set expval 127; set expvalClipped 1 }
	
	#output
	if { "$prevPbval" != $pbval } {
	    puts $fd "$tick Pb ch=$midich v=$pbval"
	    set prevPbval $pbval
	}

	if { "$prevExpval" != $expval } {
	    puts $fd "$tick Par ch=$midich c=11 v=$expval"
	    set prevExpval $expval
	}
    }

    if { $pbvalClipped } {
	cPuts "pitch clipped.\n"
    }
    if { $expvalClipped } {
	cPuts "loudness clipped.\n"
    }

    puts $fd "$tick Meta TrkEnd"
    puts $fd "TrkEnd"
    close $fd

    if { [catch {exec t2mf.exe "$tmptxtfile" "$tmpmidfile"} msg] } {
	file delete "$tmptxtfile"
	file delete "$tmpmidfile"
	tk_messageBox -icon info -message "t2mf error: $msg\n" -type ok -parent $w
	return 
    }
    
    file delete "$tmptxtfile"
    file rename -force "$tmpmidfile" "$filename"
    tk_messageBox -icon info -message "$filename created.\n" -type ok -parent $w
}






######################################################################
# Initialization
######################################################################

# draw the default horizontal scale
MVFXVizSetHorizScale $w


